from cryptography.fernet import Fernet
from dfk_commons.classes.Account import Account
from dfk_commons.classes.RPCProvider import RPCProvider
from dfk_commons.classes.TablesManager import TablesManager

def get_account(tablesManager: TablesManager, cypher_key: str, address: str, rpcProvider: RPCProvider):
    f = Fernet(cypher_key.encode())
    account_data = tablesManager.accounts.query(
            KeyConditionExpression="address_ = :address_",
            ExpressionAttributeValues={
                ":address_": address,
            })["Items"][0]
    key = account_data["key_"]
    decrypted_key = f.decrypt(key.encode()).decode()
    return Account(rpcProvider.w3.eth.account.from_key(decrypted_key), rpcProvider, account_data)

def get_account_from_private_key(private_key: str, rpcProvider: RPCProvider):
    return Account(rpcProvider.w3.eth.account.from_key(private_key), rpcProvider, None)



def attempts(prod, profession):
    if profession == "mining" or profession == "token mining":
        if prod:
            return 25
        else:
            return 1
    elif profession == "gardening":
        if prod:
            return 25
        else:
            return 1
    else:
        if prod:
            return 5
        else:
            return 1

class HeroQuestData:
    def __init__(self, prod, heroId, profession, quest_type, level, status):
        self.profession = profession
        self.heroId = heroId
        self.attempts = attempts(prod, profession)
        self.quest_type = quest_type
        self.level = level
        self.status = status

    def __str__(self):
        return str(self.__dict__)
    
    def comparation_string(self):
        return self.profession + str(self.level) + str(self.quest_type) + self.status

class QuestGroup:
    def __init__(self):
        self.heroes: list[HeroQuestData] = []
    
    def __str__(self):
        return str([str(hero) for hero in self.heroes])
    
    def addHeroToGroup(self, heroQuest: HeroQuestData):
        self.heroes.append(heroQuest)



from dfk_commons.classes.APIService import APIService
from dfk_commons.classes.Account import Account
from dfk_commons.classes.RPCProvider import RPCProvider
from dfk_commons.classes.Token import Token
from dfk_commons.abi_getters import ERC20ABI, ERC721ABI, RouterABI

def getJewelBalance(account: Account, rpcProvider: RPCProvider):
    return int(rpcProvider.w3.eth.get_balance(account.address))

def getCrystalBalance(account: Account, apiService: APIService, rpcProvider: RPCProvider):
    contract = rpcProvider.w3.eth.contract(address= apiService.tokens["Crystal"].address, abi=ERC20ABI)
    return int(contract.functions.balanceOf(account.address).call())

def getTokenAmount(account: Account, token: Token, rpcProvider: RPCProvider):
    contract = rpcProvider.w3.eth.contract(address= token.address, abi=ERC20ABI)
    return int(contract.functions.balanceOf(account.address).call())

def getTokenPriceInJewel(token: Token, apiService: APIService, rpcProvider: RPCProvider):
    RouterContract = rpcProvider.w3.eth.contract(address=apiService.contracts["Router"]["address"], abi=RouterABI)
    try:
        price = RouterContract.functions.getAmountsOut(1*(10**token.decimals), [token.address, apiService.tokens["Jewel"].address]).call()[1]
        price = price/(10**token.decimals)
    except Exception as e:
        price = 0
    return price

def heroNumber(account: Account, apiService: APIService, rpcProvider: RPCProvider):
    contract = rpcProvider.w3.eth.contract(address= apiService.contracts["Heroes"]["address"], abi=ERC721ABI)
    return int(contract.functions.balanceOf(account.address).call())

def sendJewel(account: Account, payout_address: str, amount: int, rpcProvider: RPCProvider):
    tx = {
        "from": account.address,
        "to": payout_address,
        "value": amount,
        "nonce": account.nonce,
        "chainId": rpcProvider.chainId
    }
    tx["gas"] = int(rpcProvider.w3.eth.estimate_gas(tx))
    tx["maxFeePerGas"] = rpcProvider.w3.to_wei(12, 'gwei')
    tx["maxPriorityFeePerGas"] = rpcProvider.w3.to_wei(1, "gwei")
    signed_tx = rpcProvider.w3.eth.account.sign_transaction(tx, account.key)
    hash = rpcProvider.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    hash = rpcProvider.w3.to_hex(hash)
    rpcProvider.w3.eth.wait_for_transaction_receipt(hash)

def sendToken(account: Account, payout_address, amount, token: Token, rpcProvider: RPCProvider):
    contract = rpcProvider.w3.eth.contract(address= token.address, abi=ERC20ABI)
    tx = contract.functions.transfer(
        payout_address,
        amount,
    ).build_transaction({
        "from": account.address,
        "nonce": account.nonce,
    })
    signed_tx = rpcProvider.w3.eth.account.sign_transaction(tx, account.key)
    hash = rpcProvider.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    hash = rpcProvider.w3.to_hex(hash)
    rpcProvider.w3.eth.wait_for_transaction_receipt(hash)

def checkAllowance(account: Account, token: Token, address, abi, rpcProvider: RPCProvider):
    contract = rpcProvider.w3.eth.contract(address= token.address, abi=abi)
    if int(contract.functions.allowance(account.address, address).call()) == 0:
        return True
    else: 
        return False
    
def addAllowance(account: Account, token: Token, address, abi, rpcProvider: RPCProvider):
    contract = rpcProvider.w3.eth.contract(address= token.address, abi=abi)
    tx = contract.functions.approve(address, 115792089237316195423570985008687907853269984665640564039457584007913129639935).build_transaction({
        "from": account.address,
        "nonce": account.nonce,
    })
    signed_tx = rpcProvider.w3.eth.account.sign_transaction(tx, account.key)
    hash = rpcProvider.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    hash = rpcProvider.w3.to_hex(hash)
    rpcProvider.w3.eth.wait_for_transaction_receipt(hash)


from dfk_commons.classes.Account import Account
from dfk_commons.classes.APIService import APIService
from dfk_commons.classes.RPCProvider import RPCProvider
from dfk_commons.abi_getters import HeroSaleABI

def buyHero(account: Account, heroId: int, heroPrice: int, apiService: APIService, rpcProvider: RPCProvider):
    HeroSaleContract = rpcProvider.w3.eth.contract(address=apiService.contracts["HeroSale"]["address"], abi=HeroSaleABI)
    tx = HeroSaleContract.functions.bid(heroId, heroPrice).build_transaction({
        "from": account.address,
        "nonce": account.nonce,
    })
    signed_tx = rpcProvider.w3.eth.account.sign_transaction(tx, account.key)
    hash = rpcProvider.w3.eth.send_raw_transaction(signed_tx.rawTransaction)
    hash = rpcProvider.w3.to_hex(hash)
    rpcProvider.w3.eth.wait_for_transaction_receipt(hash)


class Token:
    def __init__(self, item, chain, address, decimals, event_filter) -> None:
        self.name = item
        self.chain = chain
        self.address = address
        self.decimals = decimals
        self.event_filter = event_filter
    


from web3 import Web3
from web3.middleware import geth_poa_middleware
import requests

from dfk_commons.classes.RPCProvider import RPCProvider

def try_rpc(rpc, logger):
    try:
        session = requests.Session()
        if "id" in rpc and "password" in rpc:
            session.auth = (rpc["id"], rpc["password"])
        w3 = Web3(Web3.HTTPProvider(rpc["url"], session=session))
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        w3.client_version
        w3.eth.get_block("latest")
        return w3
    except Exception as e:
        logger.info(f"RPC {rpc['url']} failed: {e}")
    return False


def get_rpc_provider(chain, disabled_rpc_list, logger):
    if chain == "dfk":
        rpc_list = [
            {
                "url":"https://subnets.avax.network/defi-kingdoms/dfk-chain/rpc",
            },
            {
                "url":"https://avax-pokt.nodies.app/ext/bc/q2aTwKuyzgs8pynF7UXBZCU7DejbZbZ6EUyHr3JQzYgwNPUPi/rpc",
            },
            {
                "url":"https://dfkchain.api.onfinality.io/public",
            },
        ]
        chainId= 53935
    elif chain == "klay":
        rpc_list = [
            {
                "url":"https://public-en-cypress.klaytn.net",
            },
            {
                "url":"https://klaytn-pokt.nodies.app",
            },
            {
                "url":"https://klaytn.api.onfinality.io/public",
            },
        ]
        chainId = 8217

    for rpc in rpc_list:
        if rpc["url"] in disabled_rpc_list:
            continue
        w3 = try_rpc(rpc, logger)
        if w3:
            logger.info(f"Using RPC: {rpc['url']}")
            return RPCProvider(chain, w3, rpc["url"], chainId)
        
    raise Exception("No RPC available")
